//*****************************************************************************
// Projeto Ring Comm - Comunicação em anel usando RS-232
//
// Desenvolvido por Eng.: Márcio José Soares
// 
// Microcontrolador: Arduino Uno
// Sensores        : nenhum
// Extras          : Adaptador de nível TTL/RS-232 e LCD KeyPad Shield 
//
// Compilador      : Arduino-cli 0.24.0
// IDE             : Kate (Linux)
//
//
// Compilar: ~/Arduino-cli/bin/arduino-cli compile --fqbn arduino:avr:uno Ring_comm.ino 
// Upload  : ~/Arduino-cli/bin/arduino-cli upload -v -p /dev/ttyXXXX --fqbn arduino:avr:uno Ring_comm.ino 
//
//
// Pinos utilizados:
//
// 
// Últimas alterações:
//	em 19/05/2023:
//		- criado esse sketch
//
//  em 23/05/2023:
//		- inicio dos testes com STM32 e PIC 
//		- inserido saída por timeout
//		- inserido verificação de erro no nr do dispositivo
//        para evitar mensagem se repetindo
//
//  em 25/05/2023:
//		- testes para primeira versão ok!!!
//      - divulgando no site e no canal!
//*****************************************************************************

//*****************************************************************************
// Arquivos incluídos no módulo
//*****************************************************************************
#include <LiquidCrystal.h>

//*****************************************************************************
// Definições importantes do módulo
//*****************************************************************************

//*****************************************************************************
// Pinos de controle
#define pinRs			8			// LCD
#define pinEn			9
#define pinD4			4
#define pinD5			5
#define pinD6           6
#define pinD7           7
#define pinBKLT			10
#define LEDLive			13

#define pinBTNs			A0			// botões

//*****************************************************************************
// Valores para captura dos botões analógicos
#define noTerm			0			// 0 = não retorna nenhum terminal
#define Term_1			1			// msg para term 1
#define Term_3			3			// msg para term 3
#define Term_4			4           // msg para term 4

//*****************************************************************************
// Definições diversas
#define TRUE 			1
#define FALSE 			0
#define CR				0x0D
#define LF				0x0A
#define ACK				0x06

#define LEDLive 		13					// pino do LED Live

//*******************************************************************************
// Definições para uso do protocolo
#define HEADER			'@'					// header do protocolo
#define NR_DISP			0x02				// nr do dispositivo na rede!

#define POSHEADER		0x00				// posições e tamanho final do buffer
#define POSDISPS		0x01
#define POSQTBYTES		0x02
#define POSDATA			0x03

#define TAMDATA			0x0E				// apenas 14 bytes - LCD
#define TAMBUF			(0x03 + TAMDATA + 0x01)

//*****************************************************************************
// Definições de debug
//#define __DEBUG__

//*****************************************************************************
// Definições para uso do timer 1
#define __ACTIVE_TIMER__

//*****************************************************************************
// Instâncias do módulo
//*****************************************************************************
LiquidCrystal mylcd(pinRs, pinEn, pinD4, pinD5, pinD6, pinD7);

//*****************************************************************************
// Variáveis globais do módulo
//*****************************************************************************
#ifdef __DEBUG__
volatile uint8_t debug;
#endif

char TxRxbuf[TAMBUF];           			//buffer de envio/recepção
char p[15];
char RxTerm[3];							    // var para inserir dados recebidos do terminal
char dados[TAMDATA];
uint8_t mhigh, mlow;	
volatile uint8_t autoriza = 0;
volatile uint8_t counter = 0;
volatile uint8_t prima = 0;
volatile uint8_t envio = 0;
uint16_t timeLED = 0;
uint16_t timekey = 0;
uint8_t toTerm = 0;
uint8_t timeout_aut = 0;
uint16_t mytimeout = 0;
uint16_t msgcounter = 0;

//*****************************************************************************
// Constantes do módulo
//*****************************************************************************
const int LEDTIME = 250;                  	// ms  - 250 x 2ms  = 500ms
const int TIMER2s  = 1000;					// 2ms x 1000 = 2s
const int BOUNCE  = 300;					// 50ms de bounce nas teclas

//*****************************************************************************
// Declara funções presentes no módulo
//*****************************************************************************
uint8_t getBTNs(void);
void serialEvent(void);
byte valida(char dadosTerm[], char data[]);
void bin_hex(byte binchar, byte *high_char, byte *low_char);

//*****************************************************************************
// Configurações
//*****************************************************************************
void setup(){
	
	Serial.begin(9600);						// inicia porta serial
	pinMode(pinBKLT, OUTPUT);				// pino do backlight
	pinMode(LEDLive, OUTPUT);
	
	delay(500);								// aguarda dispositivos iniciar
	mylcd.begin(16,2);						// inicia LCD
	digitalWrite(pinBKLT, HIGH);			// liga o back light
	digitalWrite(LEDLive,LOW);
	
	mylcd.clear();							// apaga LCD
	mylcd.setCursor(2,0);					// posiciona para msg de abertura
	mylcd.print("BEM-VINDO AO");			//
	mylcd.setCursor(2,1);
	mylcd.print("LAB do Arne");
	
	delay(3000);							// aguarda
	
	mylcd.clear();							// apaga LCD
	mylcd.print("T>   Q>   C>    ");		// escreve na 1° linha
	mylcd.setCursor(0,1);					// muda para linha 2
	mylcd.print("->              ");
	
#ifdef __ACTIVE_TIMER__
    //--------------------------------------------------------------
    // configura timer 1
    // cristal externo a 16MHz, então clock interno a 16MHz
    // 16MHz / 1024 (prescale) = 15,625 khz => T=1/15,625kHz = 64 us
    // 64us x 32 = 2.048ms
    // então TCNT1 = 65536 - 32 = 65504 (FFE0H)

    TCCR1A = 0;                                             //configura timer 1 para operação normal
    TCCR1B = 0;                                             //limpa registrador
    TCCR1B |= (1<<CS10)|(1 << CS12);                        // configura prescaler para 1024: CS12 = 1 e CS10 = 1

    TCNT1 = 0xFFE0;                                         // incia timer com valor para que estouro ocorra em 1ms
                                                            // 65536-(16MHz/1024/0,5kHz) = 65504 = 0xFFE0
  
    TIMSK1 |= (1 << TOIE1);                                 // habilita a interrupção do TIMER1
#endif
	
}

//*****************************************************************************
// Função principal
//*****************************************************************************
void loop(){
	
	if(Serial.available() > 0){				// se tem dados, recolhe
         serialEvent();
	}
	else{
		if(autoriza){							// tem dado??
			if(valida(&RxTerm[0], &dados[0])){  // testa protocolo - valida!
				envio = 1;						// avisa que vai enviar
				bin_hex(RxTerm[0],&mhigh,&mlow);
				mylcd.setCursor(2,0);			// cursor na linha 1, coluna 2
				mylcd.write(mhigh);
				mylcd.write(mlow);
				
				bin_hex(RxTerm[1],&mhigh,&mlow);
				mylcd.setCursor(7,0);			// cursor na linha 1, coluna 7
				mylcd.write(mhigh);
				mylcd.write(mlow);
					
				msgcounter++;
				if(msgcounter > 999)
					msgcounter = 0;
				
				sprintf(p, "%03d", msgcounter);
				mylcd.setCursor(12,0);			// cursor na linha 1, coluna 12
				mylcd.print(p);
					
				mylcd.setCursor(2,1);			// cursor na linha 0, coluna 2				
				mylcd.print("              ");
				
				mylcd.setCursor(2,1);			// cursor na linha 0, coluna 2				
				mylcd.print(dados);
				
				
				TxRxbuf[0] = HEADER;			// devolve ACK
#ifdef __DEBUG__
				uint8_t aux = NR_DISP;
				aux <<= 0x04;
				aux |= RxTerm[0];
				bin_hex(aux, &mhigh, &mlow);
				TxRxbuf[1] = mhigh;
				TxRxbuf[2] = mlow;
				TxRxbuf[3] = 0x01+debug;		// quantidade de bytes
				TxRxbuf[4] = ACK;				// envia ACK
				TxRxbuf[5] = '\0';				// fecha para inserir dados
				TxRxbuf[6] = CR;
				TxRxbuf[7] = LF;
				TxRxbuf[8] = '\0';				// fecha string
#else
				TxRxbuf[1] = NR_DISP;			// número do dispositivo que envia
				TxRxbuf[1] <<= 0x04;			// posiciona no MSB
				TxRxbuf[1] |= RxTerm[0];		// número do dispositivo que receberá
				TxRxbuf[2] = 0x01;				// quantidade de bytes
				TxRxbuf[3] = ACK;				// envia ACK
				TxRxbuf[4] = '\0';				// fecha string
#endif

				for(uint8_t i=0; i<strlen(TxRxbuf);i++)	//envia dados - ACK
					Serial.write(TxRxbuf[i]);
				envio = 0;						// avisa que está pronto para receber novamente!
			}
			autoriza = 0;						// autoriza novo buffer
		}
	}
       
}

//**********************************************************************
// Trata interrupção do timer 1
//
// Entradas - vetor de interrupção
// Saídas   - nenhuma
//**********************************************************************
ISR(TIMER1_OVF_vect){
	
	if(timeout_aut){						// se autorizado contar
		mytimeout++;
		if(mytimeout >= TIMER2s){			// se chegou o tempo e não parou
			mytimeout = 0;
			timeout_aut = 0;
			prima = 0;
			envio = 0;
			autoriza = 0;
			counter = 0;
		}
	}
	
	timekey++;
	if(timekey >= LEDTIME/2){				// uma captura a cada 250ms
		timekey = 0;
		toTerm = getBTNs();	// verifica se tem chave e envia dado
		if(toTerm){
			if(!prima){						// se esta recebendo, cancela envio!
				envio = 1;					// avisa que vai iniciar transmissão
					
				TxRxbuf[0] = HEADER;		//
#ifdef __DEBUG__
				uint8_t aux = NR_DISP;
				aux <<= 0x04;
				aux |= toTerm;
				bin_hex(aux, &mhigh, &mlow);
				TxRxbuf[1] = mhigh;
				TxRxbuf[2] = mlow;
				bin_hex((uint8_t)strlen("Arduino"), &mhigh, &mlow);
				TxRxbuf[3] = mhigh;	// quantidade de caracteres
				TxRxbuf[4] = mlow;
				TxRxbuf[5] = '\0';
				strcat(TxRxbuf,"Arduino");
				p[0] = CR;
				p[1] = LF;
				p[2] = '\0';
				strcat(TxRxbuf, p);
#else
				TxRxbuf[1] = NR_DISP;					// número do dispositivo que envia
				TxRxbuf[1] <<= 0x04;					// posiciona no MSB
				TxRxbuf[1] |= toTerm;					// número do dispositivo que receberá
				TxRxbuf[2] = (char)strlen("Arduino");   // quantidade de caracteres
				TxRxbuf[3] = '\0';						// fecha para inserir dados
				strcat(TxRxbuf,"Arduino");				// insere MSG
#endif

				for(uint8_t i=0; i<strlen(TxRxbuf);i++)	//envia dados
					Serial.write(TxRxbuf[i]);
				envio = 0;								// avisa que está pronto para receber novamente!
			}
		}
	}
	
		
    timeLED++;
    if(timeLED >= LEDTIME){                         	//trata o LED a cada 500ms
        timeLED = 0;
        digitalWrite(LEDLive, digitalRead(LEDLive) ^ 1);
    }
    
    TCNT1 = 0xFFE0;                                 	//reinicia o TIMER1
    
}

//*******************************************************************************
// getBTNs - lê chaves
//
// Entradas - nenhuma
// Saídas   - nr do terminal a enviar
//*******************************************************************************
uint8_t getBTNs(void){

	int adc_key_in = analogRead(0);
		
	if (adc_key_in > 1000) return noTerm;//btnNONE;   	 // 1023
	if (adc_key_in < 50)   return Term_4;//btnRIGHT;     // 0
	if (adc_key_in < 195)  return Term_1;//btnUP;        // 99
	if (adc_key_in < 380)  return Term_3;//btnDOWN;      // 255
	if (adc_key_in < 555)  return Term_4;//btnLEFT;      // 409
	if (adc_key_in < 790)  return noTerm;//btnSELECT;
	
	return(0);

}

//*****************************************************************************
// Função serialEvent - trata dado recebido via usart
//
// Entradas - nenhuma
// Saiudas  - nenhuma
//*****************************************************************************
void serialEvent(void){
	
    char c = 0;
   
	if(!autoriza && !envio){
		c = Serial.read();

		if(!prima){									// é o primeiro caracter?
			if(c == HEADER){						// é header?
				prima = 1;
				TxRxbuf[counter++] = c;				// guarda
				timeout_aut = 0;
				mytimeout = 0;
				timeout_aut = 1;
			}
			else{
				prima = 0;
				counter = 0;
				autoriza = 0;
				timeout_aut = 0;
				mytimeout = 0;
			}
		}
		else{										// já recebeu header
			if(counter == 1){						// agora recebe nr do dispositivos
				TxRxbuf[counter++] = c;				// apenas guarda
				timeout_aut = 0;
				mytimeout = 0;
				timeout_aut = 1;
			}
			else if(counter == 2){					// agora recebe quantidade de caracteres de dados
				if(c >= 1 && c<=TAMDATA){			// quantidade dentro da faixa?
					TxRxbuf[counter++] = c;			// guarda
					timeout_aut = 0;
					mytimeout = 0;
					timeout_aut = 1;
				}
				else{								// fora da faixa!
					prima = 0;
					autoriza = 0;
					counter = 0;
					timeout_aut = 0;
					mytimeout = 0;
				}
			}
			else{											// preparado para receber dados
				TxRxbuf[counter++] = c;						// recebe os dados
				timeout_aut = 0;
				mytimeout = 0;
				timeout_aut = 1;
				if((counter - 3) >= TxRxbuf[POSQTBYTES]){	// se passou
					autoriza = 1;
					prima = 0;
					counter = 0;
					timeout_aut = 0;
					mytimeout = 0;
				}
			}
		}
	}
}

//*******************************************************************************
// Função valida - testa buffer recebido e toma ação
//                 se dados são para este terminal, executa!
//                 do contrário, passa a frente!
//
// Entradas - buffer para inserir resposta (passagem por referência)
// Saídas   - 1 se dados são para este terminal; 0 do contrário
//*******************************************************************************
byte valida(char dadosTerm[], char data[]){

	uint8_t ret = 0;
	uint8_t nrdisp = (TxRxbuf[POSDISPS] & 0x0F);

	if(nrdisp == NR_DISP){									 		// se dado é para este dispositivo
		if(TxRxbuf[POSQTBYTES] != 0x01 && TxRxbuf[POSDATA] != ACK){	// se não é retorno de ACK
			dadosTerm[0] = ((TxRxbuf[POSDISPS] & 0xF0) >> 0x04); 	// pega sender
			dadosTerm[1] = TxRxbuf[POSQTBYTES];						// pega quantidade de bytes enviados
			for(uint8_t i=0;i<TxRxbuf[POSQTBYTES];i++){				// pega dados enviados
				data[i] = TxRxbuf[POSDATA+i];						// transfere
			}
			uint8_t aux = TxRxbuf[POSQTBYTES];
			data[aux] = '\0';										// fecha string
			ret = 0x01;												// retorna válido
		}
	}
	else{															// dado não é desse dispositivo
		if((((TxRxbuf[POSDISPS] & 0xF0) >> 0x04) != NR_DISP) &&
		   (((TxRxbuf[POSDISPS] & 0xF0) >> 0x04) != 0)){		// se nr do sender não é o mesmo, do contrário morre aqui
			for(uint8_t i=0;i<(3+TxRxbuf[POSQTBYTES]);i++)			// envia dado para próximo dispositivo
				Serial.write(TxRxbuf[i]);							// dispara envio
		}
	}

	return(ret);
}

//********************************************************************************
// Função para converter binário para HEX (saída em formato ASCII)
//
// Entradas - valor a converter, dois ponteiros para retorno dos nibbles
// Saídas   - nenhuma, valores retornados por referência
//********************************************************************************
void bin_hex(byte binchar, byte *high_char, byte *low_char){

	byte high_nibble, low_nibble;

	high_nibble = (binchar & 0xF0) / 16;
	if(high_nibble > 0x09)
		*high_char = high_nibble + 0x37;
	else
		*high_char = high_nibble + 0x30;

	low_nibble = (binchar & 0x0F);

	if(low_nibble > 0x09)
		*low_char = low_nibble + 0x37;
	else
		*low_char = low_nibble + 0x30;

}
